/* validate L(#)-convex function through Hessian on all points inside domain
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"

/* returns the minimum value of func(x)    */
/* INPUT  func : pointer to the function */
/* INPUT  p    : test point in domain */
/* WORK   a    : working matrix */
static int validate(int dim, double func(int,int*), int p[], double **a) {
    make_hesse_lg(dim, func, p, a);
    return lgconv_validate_hesse(dim, a);
}

/* decide if the function is L#-convex or not */
int lgconv_validate_h(int dim, double func(int,int*),
                      int lower[], int upper[]) {
    int ret = TRUE, step;
    int *p = ivector_new(dim);
    double **a = fmatrix_new(dim, dim);

    ivector_copy(dim, p, lower);

    step = upper[0]-lower[0];
    while ((step & (step-1)) != 0) step &= step - 1;

    for (step--; ret && step>=1; step/=2) {
        do {
#ifdef DEBUG_DISP
            printf("p=");ivector_print(dim, p);
            printf("\n");
#endif /* DEBUG_DISP */
            if (!validate(dim, func, p, a)) {
                ret = FALSE;
                printf("p=");ivector_print(dim, p);
                printf(", step=%d\n", step);
                printf("hesse=");fmatrix_print(dim, a);
                printf("\n");
            }
        } while (ret && next(dim, p, lower, upper, step));
    }
    free(a);
    free(p);
    return ret;
}
