/* test validate L(#)-convex function
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "laminar.h"

/* f1(x) = 2*x0^2 + 3*x1^2 - 2*x0*x1 */
/* 2ǽ񤯤 Hessian  [2, -1; -1, 3] */
static double f1(int dim, int x[]) {
    double r = 0;

    assert(dim == 2);  /* check dimension */
    r +=  2 * x[0] * x[0];
    r +=  3 * x[1] * x[1];
    r += -2 * x[0] * x[1];
    return r;
}

/*  f1() Ʊ */
static double f2(int dim, int x[]) {
    if (0<=x[0] && x[0]<5 && 0<=x[1] && x[1]<5) return 100;
    return f1(dim, x);
}

/* ط M#-convex */
static double f3(int dim, int x[]) {
    static int v[] = {
        2,40,90, 1,2,   -1,/* y=(x1+x2),    f += 2y^2+40y+233 */
        50,-2,0, 1,     -1,/* y=(x1),       f += 50y^2-2y */
        3,20,0,  1,2,3, -1,/* y=(x1+x2+x3), f += 3y^2+20y */
        1,1,0,   0,     -1,/* y=(x0),       f += y^2 + y */
        0,0,0,             /* end mark */
    };
    return ilaminar2(dim, x, v);
}


/* LǤϤʤ (M + M) */
static double f4(int dim, int x[]) {
    double r = f3(dim, x);

    assert(dim == 4);  /* check dimension */
    r +=  2 * x[0] * x[0];
    r +=  3 * x[1] * x[1];
    r +=  2 * (x[2]+5) * (x[2]+5);
    r +=  5 * (x[3]-5) * (x[3]-5);
    return r;
}

static int test(int dim, double f(int,int*), char *f_name,
          int lower[], int upper[]) {
    int ret;

    counter_reset();
    ret = lgconv_validate_h(dim, f, lower, upper);
    if (ret) printf("%s(x) is     an L#-convex func.", f_name);
    else     printf("%s(x) is NOT an L#-convex func.", f_name);
    counter_print();
    printf("\n");
    return ret;
}

/* LȽι®С */
/* ƤʤȤĤ뤿 IFF/FW ǺŬƤߤ */
/* ְäŬμϤå */
static void test2(int dim, double f(int,int*), char *f_name, int *init) {
    int i;

    int *x     = ivector_new(dim);
    int *lower = ivector_new(dim);
    int *upper = ivector_new(dim);
    ivector_copy(dim, x, init);

    lgconv_minimize_FW(dim, f, x);
    ivector_copy(dim, lower, x);
    ivector_copy(dim, upper, x);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    test(dim, f, f_name, lower, upper);

    lgconv_minimize_IFF(dim, f, x);
    ivector_copy(dim, lower, x);
    ivector_copy(dim, upper, x);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    test(dim, f, f_name, lower, upper);
}

void test_lconv_valid(const int dim) {
    int *lower = ivector_new(10 /* dim */,-10, -10, -10, -10);
    int *upper = ivector_new(10 /* dim */, 10,  10,  10,  10);

    test2(4, f4, "f4", lower);

    test(4, f4, "f4", lower, upper);

    /*return;*/
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    lower[1]++;
    test(2, f2, "f2", lower, upper);
    test(2, f1, "f1", lower, upper);
}
