#include <stdio.h>
#include <stdlib.h>
#include "odicon.h"
#include "quadratic.h"
#include "laminar.h"


/* L# */
static double f1(int dim, int x[]) {
    static double **a;
    if (a == NULL) {
        int i, j;
        double z[4][4] = {{ 7,-3,-2,-1},
                          {-3, 9,-3,-2},
                          {-2,-3, 8,-3},
                          {-1,-2,-3, 6}};
        a = fmatrix_new(4,4);
        for (i=0; i<4; i++) for (j=0; j<4; j++) a[i][j]=z[i][j];
    }
    return quadratic(4,x,a);
}

/* M, not M# */
static double f2(int dim, int x[]) {
    static double **a;
    if (a == NULL) {
        int i, j;
        double z[4][4] = {{4,3,2,1},
                          {3,6,3,2},
                          {2,3,6,3},
                          {1,2,3,4}};
        a = fmatrix_new(4,4);
        for (i=0; i<4; i++) for (j=0; j<4; j++) a[i][j]=z[i][j];
    }
    return quadratic(4,x,a);
}


/* M and M# */
struct laminar_st *laminar_inst;
static double laminar(int dim, int x[]) {
    return ilaminartt(dim, x, laminar_inst);
}


double **hesse, *liner, constant;
static double g(int dim, int x[]) {
    return quadratic3(dim, x, hesse, liner, constant);
}

static void nomalize(const int dim, double f(int,int*), char *func_name) {
    int i, step = 10;
    int *lower = ivector_new(dim);
    int *upper = ivector_new(dim);
    int *x     = ivector_new(dim);
    for (i=0; i<dim; i++) { x[i] = 0; lower[i] = -100; upper[i] = 100; }

    hesse = fmatrix_new(dim,dim);
    liner = fvector_new(dim);

    make_hesse_lg(dim, f, x, hesse);
    make_liner_const(dim, f, x, hesse, liner, &constant);

    printf("%s is ", func_name);
    if ( lconv_validate_hesse(dim, hesse)) printf("L-convex, ");
    if (lgconv_validate_hesse(dim, hesse)) printf("L#-convex, ");
    if ( mconv_validate_hesse(dim, hesse)) printf("M-convex, ");
    if (mgconv_validate_hesse(dim, hesse)) printf("M#-convex, ");
    printf("\n A:"); fmatrix_print(dim, hesse);
    printf("\n B:"); fvector_print(dim, liner);
    printf("\n C:[%3g]\n", constant);
    
    ivector_copy(dim, x, lower);
    do {
        if (fcmp(f(dim,x)/10000, g(dim,x)/10000) != 0) {
            printf("%g - %g = %g\n", f(dim,x), g(dim,x), f(dim,x)-g(dim,x));
        }
    } while (next(dim, x, lower, upper, step));

    free(lower); free(upper); free(x); free(hesse); free(liner);
    printf("\n");
}


static void test_lg(const int dim, int *x, double f(int,int*), char *func_name) {
    double **a = fmatrix_new(dim, dim);

    printf("test_lg\n");
    make_hesse_lg(dim, f, x, a);
    fmatrix_print(dim, a);  printf("\n");
    if (lgconv_validate_hesse(dim, a)) {
        printf("%s(x) is an L#-convex func.\n\n", func_name);
    } else {
        printf("%s(x) is NOT an L#-convex func.\n\n", func_name);
    }
    free(a);
}

static void test_mg(const int dim, int *x, double f(int,int*), char *func_name) {
    double **a = fmatrix_new(dim, dim);

    printf("test_mg\n");
    make_hesse_mg(dim, f, x, a);
    fmatrix_print(dim, a);  printf("\n");
    if (mgconv_validate_hesse(dim, a)) {
        printf("%s(x) is an M#-convex func.\n\n", func_name);
    } else {
        printf("%s(x) is NOT an M#-convex func.\n\n", func_name);
    }
    free(a);
}

static void test_m(const int dim, int *x, double f(int,int*), char *func_name) {
    double **a = fmatrix_new(dim, dim);

    printf("test_m\n");
    make_hesse_m(dim, f, x, a);
    fmatrix_print(dim, a);  printf("\n");
    if (mconv_validate_hesse(dim, a)) {
        printf("%s(x) is an M-convex func.\n\n", func_name);
    } else {
        printf("%s(x) is NOT an M-convex func.\n\n", func_name);
    }
    free(a);
}

void test_quadratic(const int dim) {
    int i;
    int *x = ivector_new(dim);
    
    for (i=0; i<4; i++) x[dim] = 0;
    laminar_inst = laminartt_make(dim);

    nomalize(4, f1, "f1");
    nomalize(4, f2, "f2");
    nomalize(dim, laminar, "laminar");

    test_lg(4, x, f1, "f1");
    test_mg(4, x, f2, "f2");
    test_mg(dim, x, laminar, "laminar");

    free(x);
}
